from typing import Sequence, Tuple

import ntcore
from wpimath import geometry

import magicbot


class BasicComponent:
    @magicbot.feedback
    def get_number(self):
        return 0

    @magicbot.feedback
    def get_ints(self):
        return (0,)

    @magicbot.feedback
    def get_floats(self):
        return (0.0, 0)

    def execute(self):
        pass


class TypeHintedComponent:
    @magicbot.feedback
    def get_rotation(self) -> geometry.Rotation2d:
        return geometry.Rotation2d()

    @magicbot.feedback
    def get_rotation_array(self) -> Sequence[geometry.Rotation2d]:
        return [geometry.Rotation2d()]

    @magicbot.feedback
    def get_rotation_2_tuple(self) -> Tuple[geometry.Rotation2d, geometry.Rotation2d]:
        return (geometry.Rotation2d(), geometry.Rotation2d())

    @magicbot.feedback
    def get_int(self) -> int:
        return 0

    @magicbot.feedback
    def get_float(self) -> float:
        return 0.5

    @magicbot.feedback
    def get_ints(self) -> Sequence[int]:
        return (0,)

    @magicbot.feedback
    def get_empty_strings(self) -> Sequence[str]:
        return ()

    def execute(self):
        pass


class Robot(magicbot.MagicRobot):
    basic: BasicComponent
    type_hinted: TypeHintedComponent

    def createObjects(self):
        pass


def test_feedbacks_with_type_hints():
    robot = Robot()
    robot.robotInit()
    nt = ntcore.NetworkTableInstance.getDefault().getTable("components")

    robot._do_periodics()

    for name, type_str, value in (
        ("basic/number", "double", 0.0),
        ("basic/ints", "int[]", [0]),
        ("basic/floats", "double[]", [0.0, 0.0]),
        ("type_hinted/int", "int", 0),
        ("type_hinted/float", "double", 0.5),
        ("type_hinted/ints", "int[]", [0]),
        ("type_hinted/empty_strings", "string[]", []),
    ):
        topic = nt.getTopic(name)
        assert topic.getTypeString() == type_str
        assert topic.genericSubscribe().get().value() == value

    for name, value in [
        ("type_hinted/rotation", geometry.Rotation2d()),
    ]:
        struct_type = type(value)
        assert nt.getTopic(name).getTypeString() == f"struct:{struct_type.__name__}"
        topic = nt.getStructTopic(name, struct_type)
        assert topic.subscribe(None).get() == value

    for name, struct_type, value in (
        ("type_hinted/rotation_array", geometry.Rotation2d, [geometry.Rotation2d()]),
        (
            "type_hinted/rotation_2_tuple",
            geometry.Rotation2d,
            [geometry.Rotation2d(), geometry.Rotation2d()],
        ),
    ):
        assert nt.getTopic(name).getTypeString() == f"struct:{struct_type.__name__}[]"
        topic = nt.getStructArrayTopic(name, struct_type)
        assert topic.subscribe([]).get() == value



import collections.abc
import functools
import inspect
import typing
import warnings
from typing import Callable, Generic, Optional, TypeVar, Union, overload
from collections.abc import Sequence

import ntcore
from ntcore import NetworkTableInstance
from ntcore.types import ValueT


class StructSerializable(typing.Protocol):
    """Any type that is a wpiutil.wpistruct."""

    WPIStruct: typing.ClassVar


T = TypeVar("T")
V = TypeVar("V", bound=Union[ValueT, StructSerializable, Sequence[StructSerializable]])


class tunable(Generic[V]):
    """
    This allows you to define simple properties that allow you to easily
    communicate with other programs via NetworkTables.

    The following example will define a NetworkTable variable at
    ``/components/my_component/foo``::

        class MyRobot(magicbot.MagicRobot):

            my_component: MyComponent

        ...

        from magicbot import tunable

        class MyComponent:

            # define the tunable property
            foo = tunable(True)

            def execute(self):

                # set the variable
                self.foo = True

                # get the variable
                foo = self.foo

    The key of the NetworkTables variable will vary based on what kind of
    object the decorated method belongs to:

    * A component: ``/components/COMPONENTNAME/VARNAME``
    * An autonomous mode: ``/autonomous/MODENAME/VARNAME``
    * Your main robot class: ``/robot/VARNAME``

    .. note:: When executing unit tests on objects that create tunables,
              you will want to use setup_tunables to set the object up.
              In normal usage, MagicRobot does this for you, so you don't
              have to do anything special.

    .. versionchanged:: 2024.1.0
       Added support for WPILib Struct serializable types.
       Integer defaults now create integer topics instead of double topics.
    """

    # the way this works is we use a special class to indicate that it
    # is a tunable, and MagicRobot adds _ntattr and _global_table variables
    # to the class property

    # The tricky bit is that you need to do late binding on these, because
    # the networktables key is not known when the object is created. Instead,
    # the name of the key is related to the name of the variable name in the
    # robot class

    __slots__ = (
        "_ntdefault",
        "_ntsubtable",
        "_ntwritedefault",
        # "__doc__",
        "__orig_class__",
        "_topic_type",
        "_nt",
    )

    def __init__(
        self,
        default: V,
        *,
        writeDefault: bool = True,
        subtable: Optional[str] = None,
        doc=None,
    ) -> None:
        if doc is not None:
            warnings.warn("tunable no longer uses the doc argument", stacklevel=2)

        self._ntdefault = default
        self._ntsubtable = subtable
        self._ntwritedefault = writeDefault
        # self.__doc__ = doc

        # Defer checks for empty sequences to check type hints.
        # Report errors here when we can so the error points to the tunable line.
        if default or not isinstance(default, collections.abc.Sequence):
            topic_type = _get_topic_type_for_value(default)
            if topic_type is None:
                checked_type: type = type(default)
                raise TypeError(
                    f"tunable is not publishable to NetworkTables, type: {checked_type.__name__}"
                )
            self._topic_type = topic_type

    def __set_name__(self, owner: type, name: str) -> None:
        type_hint: Optional[type] = None
        # __orig_class__ is set after __init__, check it here.
        orig_class = getattr(self, "__orig_class__", None)
        if orig_class is not None:
            # Accept field = tunable[Sequence[int]]([])
            type_hint = typing.get_args(orig_class)[0]
        else:
            type_hint = typing.get_type_hints(owner).get(name)
            origin = typing.get_origin(type_hint)
            if origin is typing.ClassVar:
                # Accept field: ClassVar[tunable[Sequence[int]]] = tunable([])
                type_hint = typing.get_args(type_hint)[0]
                origin = typing.get_origin(type_hint)
            if origin is tunable:
                # Accept field: tunable[Sequence[int]] = tunable([])
                type_hint = typing.get_args(type_hint)[0]

        if type_hint is not None:
            topic_type = _get_topic_type(type_hint)
        else:
            topic_type = _get_topic_type_for_value(self._ntdefault)

        if topic_type is None:
            checked_type: type = type_hint or type(self._ntdefault)
            raise TypeError(
                f"tunable is not publishable to NetworkTables, type: {checked_type.__name__}"
            )

        self._topic_type = topic_type

    @overload
    def __get__(self, instance: None, owner=None) -> "tunable[V]": ...

    @overload
    def __get__(self, instance, owner=None) -> V: ...

    def __get__(self, instance, owner=None):
        if instance is not None:
            return instance._tunables[self].get()
        return self

    def __set__(self, instance, value: V) -> None:
        instance._tunables[self].set(value)


def _get_topic_type_for_value(value) -> Optional[Callable[[ntcore.Topic], typing.Any]]:
    topic_type = _get_topic_type(type(value))
    # bytes and str are Sequences. They must be checked before Sequence.
    if topic_type is None and isinstance(value, collections.abc.Sequence):
        if not value:
            raise ValueError(
                f"tunable default cannot be an empty sequence, got {value}"
            )
        topic_type = _get_topic_type(Sequence[type(value[0])])  # type: ignore [misc]
    return topic_type


def setup_tunables(component, cname: str, prefix: Optional[str] = "components") -> None:
    """
    Connects the tunables on an object to NetworkTables.

    :param component:   Component object
    :param cname:       Name of component
    :param prefix:      Prefix to use, or no prefix if None

    .. note:: This is not needed in normal use, only useful
              for testing
    """

    cls = component.__class__

    if prefix is None:
        prefix = f"/{cname}"
    else:
        prefix = f"/{prefix}/{cname}"

    NetworkTables = NetworkTableInstance.getDefault()

    tunables: dict[tunable, ntcore.Topic] = {}

    for n in dir(cls):
        if n.startswith("_"):
            continue

        prop = getattr(cls, n)
        if not isinstance(prop, tunable):
            continue

        if prop._ntsubtable:
            key = f"{prefix}/{prop._ntsubtable}/{n}"
        else:
            key = f"{prefix}/{n}"

        topic = prop._topic_type(NetworkTables.getTopic(key))
        ntvalue = topic.getEntry(prop._ntdefault)
        if prop._ntwritedefault:
            ntvalue.set(prop._ntdefault)
        else:
            ntvalue.setDefault(prop._ntdefault)
        tunables[prop] = ntvalue

    component._tunables = tunables


@overload
def feedback(f: Callable[[T], V]) -> Callable[[T], V]: ...


@overload
def feedback(*, key: str) -> Callable[[Callable[[T], V]], Callable[[T], V]]: ...


def feedback(f=None, *, key: Optional[str] = None) -> Callable:
    """
    This decorator allows you to create NetworkTables values that are
    automatically updated with the return value of a method.

    ``key`` is an optional parameter, and if it is not supplied,
    the key will default to the method name with a leading ``get_`` removed.
    If the method does not start with ``get_``, the key will be the full
    name of the method.

    The key of the NetworkTables value will vary based on what kind of
    object the decorated method belongs to:

    * A component: ``/components/COMPONENTNAME/VARNAME``
    * Your main robot class: ``/robot/VARNAME``

    The NetworkTables value will be auto-updated in all modes.

    .. warning:: The function should only act as a getter, and must not
                 take any arguments (other than self).

    Example::

        from magicbot import feedback

        class MyComponent:
            navx: ...

            @feedback
            def get_angle(self) -> float:
                return self.navx.getYaw()

        class MyRobot(magicbot.MagicRobot):
            my_component: MyComponent

            ...

    In this example, the NetworkTable key is stored at
    ``/components/my_component/angle``.

    .. seealso:: :class:`~wpilib.LiveWindow` may suit your needs,
                 especially if you wish to monitor WPILib objects.

    .. versionadded:: 2018.1.0

    .. versionchanged:: 2024.1.0
       WPILib Struct serializable types are supported when the return type is type hinted.
       An ``int`` return type hint now creates an integer topic.
    """
    if f is None:
        return functools.partial(feedback, key=key)

    if not callable(f):
        raise TypeError(f"Illegal use of feedback decorator on non-callable {f!r}")
    sig = inspect.signature(f)
    name = f.__name__

    if len(sig.parameters) != 1:
        raise ValueError(
            f"{name} may not take arguments other than 'self' (must be a simple getter method)"
        )

    # Set attributes to be checked during injection
    f._magic_feedback = True
    f._magic_feedback_key = key

    return f


_topic_types = {
    bool: ntcore.BooleanTopic,
    int: ntcore.IntegerTopic,
    float: ntcore.DoubleTopic,
    str: ntcore.StringTopic,
    bytes: ntcore.RawTopic,
}
_array_topic_types = {
    bool: ntcore.BooleanArrayTopic,
    int: ntcore.IntegerArrayTopic,
    float: ntcore.DoubleArrayTopic,
    str: ntcore.StringArrayTopic,
}


def _get_topic_type(
    return_annotation,
) -> Optional[Callable[[ntcore.Topic], typing.Any]]:
    if return_annotation in _topic_types:
        return _topic_types[return_annotation]
    if hasattr(return_annotation, "WPIStruct"):
        return lambda topic: ntcore.StructTopic(topic, return_annotation)

    # Check for PEP 484 generic types
    origin = getattr(return_annotation, "__origin__", None)
    args = typing.get_args(return_annotation)
    if origin in (list, tuple, collections.abc.Sequence) and args:
        # Ensure tuples are tuple[T, ...] or homogenous
        if origin is tuple and not (
            (len(args) == 2 and args[1] is Ellipsis) or len(set(args)) == 1
        ):
            return None

        inner_type = args[0]
        if inner_type in _array_topic_types:
            return _array_topic_types[inner_type]
        if hasattr(inner_type, "WPIStruct"):
            return lambda topic: ntcore.StructArrayTopic(topic, inner_type)

    return None


def collect_feedbacks(component, cname: str, prefix: Optional[str] = "components"):
    """
    Finds all methods decorated with :func:`feedback` on an object
    and returns a list of 2-tuples (method, NetworkTables entry setter).

    .. note:: This isn't useful for normal use.
    """
    if prefix is None:
        prefix = f"/{cname}"
    else:
        prefix = f"/{prefix}/{cname}"

    nt = NetworkTableInstance.getDefault().getTable(prefix)
    feedbacks = []

    for name, method in inspect.getmembers(component, inspect.ismethod):
        if getattr(method, "_magic_feedback", False):
            key = method._magic_feedback_key
            if key is None:
                if name.startswith("get_"):
                    key = name[4:]
                else:
                    key = name

            return_annotation = typing.get_type_hints(method).get("return", None)
            if return_annotation is not None:
                topic_type = _get_topic_type(return_annotation)
            else:
                topic_type = None

            if topic_type is None:
                entry = nt.getEntry(key)
                setter = entry.setValue
            else:
                publisher = topic_type(nt.getTopic(key)).publish()
                setter = publisher.set

            feedbacks.append((method, setter))

    return feedbacks



import hal
import wpilib


class NotifierDelay:
    """Synchronizes a timing loop against interrupts from the FPGA.

    This will delay so that the next invocation of your loop happens at
    precisely the same period, assuming that your loop does not take longer
    than the specified period.

    Example::

        with NotifierDelay(0.02) as delay:
            while something:
                # do things here
                delay.wait()
    """

    def __init__(self, delay_period: float) -> None:
        """:param delay_period: The period's amount of time (in seconds)."""
        if delay_period < 0.001:
            raise ValueError("You probably don't want to delay less than 1ms!")

        # Convert the delay period to microseconds, as FPGA timestamps are microseconds
        self.delay_period = int(delay_period * 1e6)
        self._notifier = hal.initializeNotifier()[0]
        self._expiry_time = wpilib.RobotController.getFPGATime() + self.delay_period
        self._update_alarm(self._notifier)

        # wpilib.Resource._add_global_resource(self)

    def __del__(self):
        self.free()

    def __enter__(self) -> "NotifierDelay":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.free()

    def free(self) -> None:
        """Clean up the notifier.

        Do not use this object after this method is called.
        """
        handle = self._notifier
        if handle is None:
            return
        hal.stopNotifier(handle)
        hal.cleanNotifier(handle)
        self._notifier = None

    def wait(self) -> None:
        """Wait until the delay period has passed."""
        handle = self._notifier
        if handle is None:
            return
        hal.waitForNotifierAlarm(handle)
        self._expiry_time += self.delay_period
        self._update_alarm(handle)

    def _update_alarm(self, handle) -> None:
        hal.updateNotifierAlarm(handle, self._expiry_time)



from typing import ClassVar, List, Sequence

import ntcore
import pytest
from wpimath import geometry

from magicbot.magic_tunable import setup_tunables, tunable


def test_tunable() -> None:
    class Component:
        an_int = tunable(1)
        ints = tunable([0])
        floats = tunable([1.0, 2.0])
        rotation = tunable(geometry.Rotation2d())
        rotations = tunable([geometry.Rotation2d()])

    component = Component()
    setup_tunables(component, "test_tunable")
    nt = ntcore.NetworkTableInstance.getDefault().getTable("/components/test_tunable")

    for name, type_str, value in [
        ("an_int", "int", 1),
        ("ints", "int[]", [0]),
        ("floats", "double[]", [1.0, 2.0]),
    ]:
        topic = nt.getTopic(name)
        assert topic.getTypeString() == type_str
        assert topic.genericSubscribe().get().value() == value
        assert getattr(component, name) == value

    for name, value in [
        ("rotation", geometry.Rotation2d()),
    ]:
        struct_type = type(value)
        assert nt.getTopic(name).getTypeString() == f"struct:{struct_type.__name__}"
        topic = nt.getStructTopic(name, struct_type)
        assert topic.subscribe(None).get() == value
        assert getattr(component, name) == value

    for name, struct_type, value in [
        ("rotations", geometry.Rotation2d, [geometry.Rotation2d()]),
    ]:
        assert nt.getTopic(name).getTypeString() == f"struct:{struct_type.__name__}[]"
        topic = nt.getStructArrayTopic(name, struct_type)
        assert topic.subscribe([]).get() == value
        assert getattr(component, name) == value


def test_tunable_errors():
    with pytest.raises(TypeError):

        class Component:
            invalid = tunable(None)


def test_tunable_errors_with_empty_sequence():
    with pytest.raises((RuntimeError, ValueError)):

        class Component:
            empty = tunable([])


def test_type_hinted_empty_sequences() -> None:
    class Component:
        generic_seq = tunable[Sequence[int]](())
        class_var_seq: ClassVar[tunable[Sequence[int]]] = tunable(())
        inst_seq: Sequence[int] = tunable(())

        generic_typing_list = tunable[List[int]]([])
        class_var_typing_list: ClassVar[tunable[List[int]]] = tunable([])
        inst_typing_list: List[int] = tunable([])

        generic_list = tunable[list[int]]([])
        class_var_list: ClassVar[tunable[list[int]]] = tunable([])
        inst_list: list[int] = tunable([])

    component = Component()
    setup_tunables(component, "test_type_hinted_sequences")
    NetworkTables = ntcore.NetworkTableInstance.getDefault()
    nt = NetworkTables.getTable("/components/test_type_hinted_sequences")

    for name in [
        "generic_seq",
        "class_var_seq",
        "inst_seq",
        "generic_typing_list",
        "class_var_typing_list",
        "inst_typing_list",
        "generic_list",
        "class_var_list",
        "inst_list",
    ]:
        assert nt.getTopic(name).getTypeString() == "int[]"
        entry = nt.getEntry(name)
        assert entry.getIntegerArray(None) == []
        assert getattr(component, name) == []



"""
Taken from the python documentation, distributed under
that same license
"""

import collections


class OrderedClass(type):
    """
    Metaclass that stores class attributes in a member called
    'members'. If you subclass something that uses this metaclass,
    the base class members will be listed before the subclass
    """

    # TODO: this isn't required in Python 3.6 due to PEP 520

    @classmethod
    def __prepare__(metacls, name, bases, **kwds):
        return collections.OrderedDict()

    def __new__(cls, name, bases, namespace, **kwds):
        result = type.__new__(cls, name, bases, dict(namespace))

        members = []
        members.extend(getattr(result, "members", ()))

        # because of the potential of multiple inheritance/mixins, need
        # to grab members list from all bases
        for base in bases:
            members.extend(getattr(base, "members", ()))

        members.extend(namespace)

        seen = set()
        result.members = tuple(m for m in members if not (m in seen or seen.add(m)))
        return result



from typing import List, Tuple, Type, TypeVar
from unittest.mock import Mock

import magicbot


class Injectable:
    def __init__(self, num):
        self.num = num


class DumbComponent:
    def execute(self):
        pass


class Component1:
    intvar: int
    tupvar: tuple

    injectable: Injectable

    def execute(self):
        pass


class Component2:
    tupvar: tuple

    component1: Component1

    def execute(self):
        pass


class Component3:
    intvar: int

    def __init__(
        self,
        tupvar: Tuple[int, int],
        injectable: Injectable,
        component2: Component2,
    ) -> None:
        self.tuple_ = tupvar
        self.injectable_ = injectable
        self.component_2 = component2

    def execute(self):
        pass


class SimpleBot(magicbot.MagicRobot):
    intvar = 1
    tupvar = 1, 2

    component1: Component1
    component2: Component2
    component3: Component3

    def createObjects(self):
        self.injectable = Injectable(42)


class DuplicateComponent:
    var: tuple
    injectable: Injectable

    def execute(self):
        pass


class MultilevelBot(magicbot.MagicRobot):
    dup1: DuplicateComponent
    dup1_var = 1, 2

    dup2: DuplicateComponent
    dup2_var = 3, 4

    dup3: DuplicateComponent
    dup3_var = 5, 6

    def createObjects(self):
        self.injectable = Injectable(42)


class SuperComponent:
    intvar: int

    def execute(self):
        pass


class InheritedComponent(SuperComponent):
    tupvar: tuple

    def execute(self):
        pass


class InheritBot(magicbot.MagicRobot):
    component: InheritedComponent

    def createObjects(self):
        self.intvar = 1
        self.tupvar = 1, 2


class BotBase(magicbot.MagicRobot):
    component_a: DumbComponent

    def createObjects(self):
        pass


class InheritedBot(BotBase):
    component_b: DumbComponent


class TypeHintedBot(magicbot.MagicRobot):
    some_int: int = 1
    some_float: float

    component: DumbComponent

    def createObjects(self):
        self.some_float = 0.5


class TypeHintedComponent:
    injectable: Injectable
    numbers: List[float]

    some_int: int = 1
    maybe_float: float = None
    calculated_num: float

    def __init__(self):
        self.calculated_num = 1 - self.some_int
        self.numbers = [1, 2.0, 3, 5.0]

    def execute(self):
        pass


class TypeHintsBot(magicbot.MagicRobot):
    injectable: Injectable
    injectables: List[Injectable]

    component: TypeHintedComponent

    def createObjects(self):
        self.injectable = Injectable(42)
        self.injectables = [self.injectable]


R = TypeVar("R", bound=magicbot.MagicRobot)


def _make_bot(cls: Type[R]) -> R:
    bot = cls()
    bot.createObjects()
    bot._automodes = Mock()
    bot._automodes.modes = {}
    bot._create_components()
    return bot


def test_simple_annotation_inject():
    bot = _make_bot(SimpleBot)

    assert bot.component1.intvar == 1
    assert isinstance(bot.component1.injectable, Injectable)
    assert bot.component1.injectable.num == 42

    assert bot.component2.tupvar == (1, 2)
    assert bot.component2.component1 is bot.component1

    assert bot.component3.intvar == 1
    assert bot.component3.tuple_ == (1, 2)
    assert isinstance(bot.component3.injectable_, Injectable)
    assert bot.component3.injectable_.num == 42
    assert bot.component3.component_2 is bot.component2

    # Check the method hasn't been mutated
    assert str(Component3.__init__.__annotations__["return"]) == "None"


def test_multilevel_annotation_inject():
    bot = _make_bot(MultilevelBot)

    assert bot.dup1 is not bot.dup2
    assert bot.dup1.var == (1, 2)
    assert bot.dup2.var == (3, 4)
    assert bot.dup3.var == (5, 6)


def test_inherited_annotation_inject():
    bot = _make_bot(InheritBot)

    assert bot.component.tupvar == (1, 2)
    assert bot.component.intvar == 1


def test_botinherit_annotation_inject():
    bot = _make_bot(InheritedBot)

    assert isinstance(bot.component_a, DumbComponent)
    assert isinstance(bot.component_b, DumbComponent)
    assert bot.component_a is not bot.component_b


def test_typehintedbot():
    bot = _make_bot(TypeHintedBot)

    assert isinstance(bot.component, DumbComponent)
    assert bot.some_int == 1
    assert bot.some_float == 0.5


def test_typehints_inject():
    bot = _make_bot(TypeHintsBot)

    assert isinstance(bot.component, TypeHintedComponent)
    assert bot.component.some_int == 1
    assert isinstance(bot.component.injectable, Injectable)
    assert bot.component.injectable.num == 42



from typing import Any, Generic, TypeVar

V = TypeVar("V")


class will_reset_to(Generic[V]):
    """
    This marker indicates that this variable on a component will be
    reset to a default value at the very end of each control loop.

    Example usage::

        class Component:

            foo = will_reset_to(False)

            def control_fn(self):
                self.foo = True

            def execute(self):
                if self.foo:
                    # ...

                # after all components are executed, foo is reset
                # back to the default value (False)


    .. note:: This will only work for MagicRobot components

    .. warning:: This will not work on classes that set ``__slots__``.
    """

    __slots__ = ("default",)

    def __init__(self, default: V) -> None:
        self.default = default


def collect_resets(cls: type) -> dict[str, Any]:
    """
    Get all the ``will_reset_to`` variables and their values from a class.

    .. note:: This isn't useful for normal use.
    """

    result = {}

    for n in dir(cls):
        v = getattr(cls, n)
        if isinstance(v, will_reset_to):
            result[n] = v.default

    return result



from functools import partial

import wpilib


class Toggle:
    """Utility class for joystick button toggle

    Usage::

        foo = Toggle(joystick, 3)

        if foo:
            toggleFunction()

        if foo.on:
            onToggle()

        if foo.off:
            offToggle()
    """

    class _SteadyDebounce:
        """
        Similar to ButtonDebouncer, but the output stays steady for
        the given periodic_filter. E.g, if you set the period to 2
        and press the button, the value will return true for 2 seconds.

        Steady debounce will return true for the given period, allowing it to be
        used with Toggle
        """

        def __init__(self, joystick: wpilib.Joystick, button: int, period: float):
            """
            :param joystick:  Joystick object
            :type  joystick:  :class:`wpilib.Joystick`
            :param button: Number of button to retrieve
            :type  button: int
            :param period:    Period of time (in seconds) to wait before allowing new button
                              presses. Defaults to 0.5 seconds.
            :type  period:    float
            """
            self.joystick = joystick
            self.button = button

            self.debounce_period = float(period)
            self.latest = (
                -self.debounce_period
            )  # Negative latest prevents get from returning true until joystick is presed for the first time
            self.enabled = False

        def get(self):
            """
            :returns: The value of the joystick button. Once the button is pressed,
            the return value will be `True` until the time expires
            """

            now = wpilib.Timer.getFPGATimestamp()
            if now - self.latest < self.debounce_period:
                return True

            if self.joystick.getRawButton(self.button):
                self.latest = now
                return True
            else:
                return False

    def __init__(
        self, joystick: wpilib.Joystick, button: int, debounce_period: float = None
    ):
        """
        :param joystick: :class:`wpilib.Joystick` that contains the button to toggle
        :param button: Number of button that will act as toggle. Same value used in `getRawButton()`
        :param debounce_period: Period in seconds to wait before registering a new button press.
        """

        if debounce_period is not None:
            self.joystickget = Toggle._SteadyDebounce(
                joystick, button, debounce_period
            ).get
        else:
            self.joystick = joystick
            self.joystickget = partial(self.joystick.getRawButton, button)

        self.released = False
        self.toggle = False
        self.state = False

    def get(self):
        """
        :return: State of toggle
        :rtype: bool
        """
        current_state = self.joystickget()

        if current_state and not self.released:
            self.released = True
            self.toggle = not self.toggle
            self.state = not self.state  # Toggles between 1 and 0.

        elif not current_state and self.released:
            self.released = False

        return self.toggle

    @property
    def on(self):
        """
        Equates to true if toggle is in the 'on' state
        """
        self.get()
        return self.state

    @property
    def off(self):
        """
        Equates to true if toggle is in the 'off' state
        """
        self.get()
        return not self.state

    __bool__ = get



class Unit:
    """The class for all of the units here"""

    def __init__(self, base_unit, base_to_unit, unit_to_base):
        """
        Unit constructor, used as a mechanism to convert between various measurements
        :param base_unit: The instance of Unit to base conversions from. If None, then assume it is the ultimate base unit
        :param base_to_unit: A callable to convert measurements between this unit and the base unit
        :param unit_to_base: A callable to convert measurements between the base unit and this unit
        """
        self.base_unit = base_unit
        self.base_to_unit = base_to_unit
        self.unit_to_base = unit_to_base


def convert(source_unit, target_unit, value):
    """
    Convert between units, returns value in target_unit
    :param source_unit: The unit of value
    :param target_unit: The desired output unit
    :param value: The value, in source_unit, to convert
    """
    # Convert value from source_unit to the ultimate base unit
    current_unit = source_unit
    current_value = value
    while current_unit.base_unit is not None:
        current_value = current_unit.unit_to_base(current_value)
        next_unit = current_unit.base_unit
        current_unit = next_unit

    # Get the chain of conversions between target_unit and the ultimate base unit
    current_unit = target_unit
    unit_chain = []
    while current_unit.base_unit is not None:
        unit_chain.append(current_unit)
        next_unit = current_unit.base_unit
        current_unit = next_unit

    # Follow the chain of conversions back to target_unit
    for unit in reversed(unit_chain):
        current_value = unit.base_to_unit(current_value)

    # Return it!
    return current_value


# Some typical units to be used
meter = Unit(base_unit=None, base_to_unit=lambda x: None, unit_to_base=lambda x: None)
centimeter = Unit(
    base_unit=meter, base_to_unit=lambda x: x * 100, unit_to_base=lambda x: x / 100
)

foot = Unit(
    base_unit=meter,
    base_to_unit=lambda x: x / 0.3048,
    unit_to_base=lambda x: x * 0.3048,
)
inch = Unit(
    base_unit=foot, base_to_unit=lambda x: x * 12, unit_to_base=lambda x: x / 12
)
